@@ -1,35 +1,29 @@
 import numpy as np
 import astropy.units as U
 from astropy import constants as C
-from scipy.special import erf as _erf
+from scipy.special import erf
 from abc import ABCMeta, abstractmethod
 
 
-def erf(z):
-    return _erf(z.to(U.dimensionless_unscaled).value)
-
-
-class _BaseSpectrum(object):
+class _BaseSpectrum(metaclass=ABCMeta):
     """
     Abstract base class for implementions of spectral models to inherit from.
 
-    Classes inheriting from _BaseSpectrum must implement three methods:
-    `half_width`, `spectral_function` and `spectral_function_kwargs`.
+    Classes inheriting from _BaseSpectrum must implement two methods:
+    `half_width` and `spectral_function`.
 
     `half_width` should define a characteristic width for the model, measured
     from the peak to the characteristic location. Note that particles whose
     spectra within +/- 4 half-widths of the peak do not intersect the DataCube
     bandpass will be discarded to speed computation.
 
-    `spectral_function` should define the model spectrum. Arguments which
-    depend on the martini.sources.SPHSource (or derived class) properties
-    should make use of the 'spectral_function_kwargs' method. The spectrum
-    should integrate to 1, the amplitude is handled separately.
+    `spectral_function` should define the model spectrum. The spectrum should integrate to
+    1, the amplitude is handled separately.
 
-    `spectral_function_kwargs` should provide a helper to pass properties from
-    martini.source.SPHSource (or derived class) to the 'spectral_function'.
-    This is required because the source object is not accessible at class
-    initialization.
+    They may also override the function `init_spectral_function_extra_data` to make
+    information that depends on the martini.sources.SPHSource (or derived class)
+    or martini.datacube.DataCube properties available internally. This is required because
+    the source object is not accessible at class initialization.
 
     See Also
     --------
@@ -38,21 +32,22 @@ class _BaseSpectrum(object):
     DiracDeltaSpectrum
     """
 
-    __metaclass__ = ABCMeta
-
-    def __init__(self):
+    def __init__(self, spec_dtype=np.float64):
+        self.spectral_function_extra_data = None
         self.spectra = None
+        self.spec_dtype = spec_dtype
         return
 
     def init_spectra(self, source, datacube):
         """
         Pre-compute the spectrum of each particle.
 
-        The spectral model defined in 'spectral_function' is evaluated using
+        The spectral model defined in `spectral_function` is evaluated using
         the channel edges from the DataCube instance and the particle
         velocities of the SPHSource (or derived class) instance provided.
-        Additional particle properties can be accessed via the
-        'spectral_function_kwargs' helper method.
+        Initializes additional particle properties by calling
+        `init_spectral_function_extra_data` which then becomes accessible via
+        `spectral_function_extra_data`.
 
         Parameters
         ----------
@@ -65,36 +60,44 @@ class _BaseSpectrum(object):
         """
 
         channel_edges = datacube.channel_edges
-        channel_widths = np.diff(channel_edges).to(U.km * U.s ** -1)
-        vmids = source.sky_coordinates.radial_velocity
-        A = source.mHI_g * \
-            np.power(source.sky_coordinates.distance.to(U.Mpc), -2)
+        channel_widths = np.diff(channel_edges).to(U.km * U.s**-1)
+        vmids = source.skycoords.radial_velocity
+        A = source.mHI_g * np.power(source.skycoords.distance.to(U.Mpc), -2)
         MHI_Jy = (
-            U.solMass * U.Mpc ** -2 * (U.km * U.s ** -1) ** -1,
+            U.Msun * U.Mpc**-2 * (U.km * U.s**-1) ** -1,
             U.Jy,
-            lambda x: (1 / 2.36E5) * x,
-            lambda x: 2.36E5 * x
+            lambda x: (1 / 2.36e5) * x,
+            lambda x: 2.36e5 * x,
         )
-        spectral_function_kwargs = \
-            {k: np.tile(v, np.shape(channel_edges[:-1]) + (1,) * vmids.ndim).T
-             for k, v in self.spectral_function_kwargs(source).items()}
+        self.init_spectral_function_extra_data(source, datacube)
+        channel_edges_unit = channel_edges.unit
+        vmids_unit = vmids.unit
         raw_spectra = self.spectral_function(
-            np.tile(
-                channel_edges.value[:-1],
-                vmids.shape + (1,)
-            ) * channel_edges.unit,
-            np.tile(
-                channel_edges.value[1:],
-                vmids.shape + (1,)
-            ) * channel_edges.unit,
-            np.tile(
-                vmids.value,
-                np.shape(channel_edges[:-1]) + (1,) * vmids.ndim
-            ).T * vmids.unit,
-            **spectral_function_kwargs
-        ).to(U.dimensionless_unscaled).value
-        self.spectra = (A[..., np.newaxis] * raw_spectra / channel_widths)\
-            .to(U.Jy, equivalencies=[MHI_Jy])
+            (
+                np.tile(
+                    channel_edges.to_value(channel_edges_unit)[:-1], vmids.shape + (1,)
+                )
+                * channel_edges.unit
+            ).astype(self.spec_dtype),
+            (
+                np.tile(
+                    channel_edges.to_value(channel_edges_unit)[1:], vmids.shape + (1,)
+                )
+                * channel_edges.unit
+            ).astype(self.spec_dtype),
+            (
+                np.tile(
+                    vmids.to_value(vmids_unit),
+                    np.shape(channel_edges[:-1]) + (1,) * vmids.ndim,
+                ).T
+                * vmids.unit
+            ).astype(self.spec_dtype),
+        ).to_value(U.dimensionless_unscaled)
+        self.spectra = (
+            A.astype(self.spec_dtype)[..., np.newaxis]
+            * raw_spectra
+            / channel_widths.astype(self.spec_dtype)
+        ).to(U.Jy, equivalencies=[MHI_Jy])
 
         return
 
@@ -113,7 +116,7 @@ class _BaseSpectrum(object):
         pass
 
     @abstractmethod
-    def spectral_function(self, a, b, vmids, **kwargs):
+    def spectral_function(self, a, b, vmids):
         """
         Abstract method; implementation of the spectral model.
 
@@ -131,37 +134,44 @@ class _BaseSpectrum(object):
         vmids : Quantity, with dimensions of velocity
             Particle velocities along the line of sight.
 
-        **kwargs
-            See spectral_function_kwargs.
-
         See Also
         --------
-        spectral_function_kwargs
+        init_spectral_function_extra_data
         """
 
         pass
 
-    @abstractmethod
-    def spectral_function_kwargs(self, source):
+    def init_spectral_function_extra_data(self, source, datacube):
         """
-        Abstract method; helper method to pass additional arguments to the
-        spectral_function.
+        Initialize extra data needed by spectral function. Default is no extra data.
 
-        Should return a dict containing the kwarg names as keys with the values
-        to pass as associated values.
+        Derived classes should override this function, if needed, to populate the dict
+        with any information from the source that is required by the spectral_function,
+        then call super().init_spectral_function_extra_data.
 
         Parameters
         ----------
         source : martini.sources.SPHSource (or derived class) instance
-            The source object will be provided so that its attributes can be
-            accessed.
+            Source object, making particle properties available.
 
         See Also
         --------
-        GaussianSpectrum (example implementation)
+        GaussianSpectrum.init_spectral_function_extra_data (for an example with extra
+                                                            data)
         """
-
-        pass
+        if self.spectral_function_extra_data is None:
+            self.spectral_function_extra_data = dict()
+        self.spectral_function_extra_data = {
+            k: np.tile(
+                v,
+                np.shape(datacube.channel_edges[:-1])
+                + (1,) * source.skycoords.radial_velocity.ndim,
+            )
+            .astype(self.spec_dtype)
+            .T
+            for k, v in self.spectral_function_extra_data.items()
+        }
+        return
 
 
 class GaussianSpectrum(_BaseSpectrum):
@@ -187,16 +197,16 @@ class GaussianSpectrum(_BaseSpectrum):
     DiracDeltaSpectrum
     """
 
-    def __init__(self, sigma=7. * U.km * U.s ** -1):
-
+    def __init__(self, sigma=7.0 * U.km * U.s**-1, spec_dtype=np.float64):
         self.sigma_mode = sigma
-        super().__init__()
+        super().__init__(spec_dtype=spec_dtype)
 
         return
 
-    def spectral_function(self, a, b, vmids, sigma=1.0 * U.km * U.s ** -1):
+    def spectral_function(self, a, b, vmids):
         """
-        Evaluate a Gaussian integral in a channel.
+        Evaluate a Gaussian integral in a channel. Requires sigma to be available
+        from `spectral_function_extra_data`.
 
         Parameters
         ----------
@@ -209,43 +219,41 @@ class GaussianSpectrum(_BaseSpectrum):
         vmids : Quantity, with dimensions of velocity
             Particle velocities along the line of sight.
 
-        sigma : Quantity, with dimensions of velocity
-            Velocity dispersion for HI line width, either for each particle or
-            constant.
-
         Returns
         -------
         out : Quantity, dimensionless
             The evaluated spectral model.
         """
 
-        # erf strips units
-        return .5 * (
-            erf((b - vmids) / (np.sqrt(2.) * sigma)) -
-            erf((a - vmids) / (np.sqrt(2.) * sigma))
-        ) * U.dimensionless_unscaled
+        assert self.spectral_function_extra_data is not None
+        sigma = self.spectral_function_extra_data["sigma"]
 
-    def spectral_function_kwargs(self, source):
+        return 0.5 * (
+            erf((b - vmids) / (np.sqrt(2.0) * sigma))
+            - erf((a - vmids) / (np.sqrt(2.0) * sigma))
+        )
+
+    def init_spectral_function_extra_data(self, source, datacube):
         """
-        Helper function to pass particle velocity dispersions to the
-        spectral_function.
+        Helper function to expose particle velocity dispersions to `spectral_function`.
 
         Parameters
         ----------
         source : martini.sources.SPHSource (or derived class) instance
             Source object.
 
-        Returns
-        -------
-        out : dict
-            Keyword arguments for the spectral_function.
+        datacube: martini.datacube.DataCube instance
+            DataCube object.
+
         """
 
-        return {'sigma': self.half_width(source)}
+        self.spectral_function_extra_data = dict(sigma=self.half_width(source))
+        super().init_spectral_function_extra_data(source, datacube)
+        return
 
     def half_width(self, source):
         """
-        Calculate velocity dispersions from particle temperatures, or return
+        Calculate 1D velocity dispersions from particle temperatures, or return
         constant.
 
         Parameters
@@ -259,8 +267,10 @@ class GaussianSpectrum(_BaseSpectrum):
             Velocity dispersion (constant, or per particle).
         """
 
-        if self.sigma_mode == 'thermal':
-            return np.sqrt(C.k_B * source.T_g / C.m_p / 3).to(U.km * U.s ** -1)
+        if self.sigma_mode == "thermal":
+            # 3D velocity dispersion of an ideal gas is sqrt(3 * kB * T / mp)
+            # So 1D velocity dispersion is sqrt(kB * T / mp)
+            return np.sqrt(C.k_B * source.T_g / C.m_p).to(U.km * U.s**-1)
         else:
             return self.sigma_mode
 
@@ -273,8 +283,8 @@ class DiracDeltaSpectrum(_BaseSpectrum):
     velocity.
     """
 
-    def __init__(self):
-        super().__init__()
+    def __init__(self, spec_dtype=np.float64):
+        super().__init__(spec_dtype=spec_dtype)
         return
 
     def spectral_function(self, a, b, vmids):
@@ -298,24 +308,7 @@ class DiracDeltaSpectrum(_BaseSpectrum):
             The evaluated spectral model.
         """
 
-        return np.heaviside(vmids - a, 1.) * np.heaviside(b - vmids, 0.)
-
-    def spectral_function_kwargs(self, source):
-        """
-        No additional kwargs.
-
-        Parameters
-        ----------
-        source : martini.sources.SPHSource (or derived class) instance
-            Source object, making particle properties available.
-
-        Returns
-        -------
-        out : dict
-            Empty; no additional kwargs.
-        """
-
-        return dict()
+        return np.heaviside(vmids - a, 1.0) * np.heaviside(b - vmids, 0.0)
 
     def half_width(self, source):
         """
@@ -332,4 +325,4 @@ class DiracDeltaSpectrum(_BaseSpectrum):
             Velocity dispersion of 0 km/s.
         """
 
-        return 0 * U.km * U.s ** -1
+        return 0 * U.km * U.s**-1
